#include <math.h>
#include <stdio.h>
#include "km.h"

void accprim(int n,double *m,double *x,double *y,double *z,
	     double *vx,double *vy,double *vz,
	     double *ax,double *ay,double *az,double *ph)
{
  int i,l;
  double r,**r3;
  double pom,pom1,pom2;
  double gm;

  gm=ph[1];
  
  r3=dmatrix(0,n,0,n);
  for(i=1;i<=n;i++){
    r=sqrt(DSQR(x[i])+DSQR(y[i])+DSQR(z[i]));
    r3[i][0]=r*r*r;
  }
  for(i=1;i<=n;i++)
    for(l=i+1;l<=n;l++){
      r=sqrt(DSQR(x[i]-x[l])+DSQR(y[i]-y[l])+DSQR(z[i]-z[l]));
      r3[i][l]=r*r*r;
      r3[l][i]=r3[i][l];
    }
  for(i=1;i<=n;i++){
    pom=-gm*(1+m[i])/r3[i][0];
    ax[i]=pom*x[i];
    ay[i]=pom*y[i];
    az[i]=pom*z[i];
  }
  for(i=1;i<=n;i++)
    for(l=1;l<=n;l++)
      if(l!=i){
	pom1=-gm*m[l]/r3[i][l];
	pom2=-gm*m[l]/r3[l][0];
	ax[i]+=(pom1*(x[i]-x[l])+pom2*x[l]);
	ay[i]+=(pom1*(y[i]-y[l])+pom2*y[l]);
	az[i]+=(pom1*(z[i]-z[l])+pom2*z[l]);
      }
  free_dmatrix(r3,0,n,0,n);
}

void accprimrel(int n,double *m,double *x,double *y,double *z,
		double *vx,double *vy,double *vz,
		double *ax,double *ay,double *az,double *ph)
{
  int i,l;
  double *r,*v2,*rv,**r3;
  double pom,pom1,pom2;
  double c2,delta;
  double gm;

  gm=ph[1];
  
  c2=DSQR(C1);
  r=dvector(0,n);
  v2=dvector(0,n);
  rv=dvector(0,n);
  r3=dmatrix(0,n,0,n);
  for(i=1;i<=n;i++){
    r[i]=sqrt(DSQR(x[i])+DSQR(y[i])+DSQR(z[i]));
    r3[i][0]=r[i]*r[i]*r[i];
    v2[i]=DSQR(vx[i])+DSQR(vy[i])+DSQR(vz[i]);
    rv[i]=x[i]*vx[i]+y[i]*vy[i]+z[i]*vz[i];
  }
  for(i=1;i<=n;i++)
    for(l=i+1;l<=n;l++){
      delta=sqrt(DSQR(x[i]-x[l])+DSQR(y[i]-y[l])+DSQR(z[i]-z[l]));
      r3[i][l]=delta*delta*delta;
      r3[l][i]=r3[i][l];
    }
  for(i=1;i<=n;i++){
    pom=-gm*(1+m[i])/r3[i][0];
    ax[i]=pom*x[i];
    ay[i]=pom*y[i];
    az[i]=pom*z[i];
  }
  /* relativistic correction */
  pom=4*gm/c2;
  for(i=1;i<=n;i++){
    pom1=gm/r3[i][0]*(pom/r[i]-v2[i]/c2);
    pom2=pom*rv[i]/r3[i][0];
    ax[i]+=pom1*x[i]+pom2*vx[i];
    ay[i]+=pom1*y[i]+pom2*vy[i];
    az[i]+=pom1*z[i]+pom2*vz[i];
  }
  for(i=1;i<=n;i++)
    for(l=1;l<=n;l++)
      if(l!=i){
	pom1=-gm*m[l]/r3[i][l];
	pom2=-gm*m[l]/r3[l][0];
	ax[i]+=(pom1*(x[i]-x[l])+pom2*x[l]);
	ay[i]+=(pom1*(y[i]-y[l])+pom2*y[l]);
	az[i]+=(pom1*(z[i]-z[l])+pom2*z[l]);
      }
  free_dvector(r,0,n);
  free_dvector(v2,0,n);
  free_dvector(rv,0,n);
  free_dmatrix(r3,0,n,0,n);
}

void accsec(int n,double *m,double *x,double *y,double *z,int ns,
	    double *xs,double *ys,double *zs,
	    double *vxs,double *vys,double *vzs,
	    double *axs,double *ays,double *azs,double *ph,int *flag)
{
  int i,l;
  double r,**r3,rp,*rp3;
  double pom,pom1,pom2;
  double gm;

  gm=ph[1];
  
  r3=dmatrix(0,ns,0,n);
  rp3=dvector(0,n);

  for(i=1;i<=ns;i++){
    if(flag[i]==1){
      r=sqrt(DSQR(xs[i])+DSQR(ys[i])+DSQR(zs[i]));
      r3[i][0]=r*r*r;
      for(l=1;l<=n;l++){
	r=sqrt(DSQR(xs[i]-x[l])+DSQR(ys[i]-y[l])+DSQR(zs[i]-z[l]));
	r3[i][l]=r*r*r;
      }
    }
  }
  for(l=1;l<=n;l++){
    rp=sqrt(DSQR(x[l])+DSQR(y[l])+DSQR(z[l]));
    rp3[l]=rp*rp*rp;
  }
  for(i=1;i<=ns;i++){
    if(flag[i]==1){
      pom=-gm/r3[i][0];
      axs[i]=pom*xs[i];
      ays[i]=pom*ys[i];
      azs[i]=pom*zs[i];
      for(l=1;l<=n;l++){
	pom1=-gm*m[l]/r3[i][l];
	pom2=-gm*m[l]/rp3[l];
	axs[i]+=(pom1*(xs[i]-x[l])+pom2*x[l]);
	ays[i]+=(pom1*(ys[i]-y[l])+pom2*y[l]);
	azs[i]+=(pom1*(zs[i]-z[l])+pom2*z[l]);
      }
    } else {
      axs[i]=ays[i]=azs[i]=0;
    }
  }
  
  free_dmatrix(r3,0,ns,0,n);
  free_dvector(rp3,0,n);
}

void accsecrel(int n,double *m,double *x,double *y,double *z,int ns,
	       double *xs,double *ys,double *zs,
	       double *vxs,double *vys,double *vzs,
	       double *axs,double *ays,double *azs,double *ph,int *flag)
{
  int i,l;
  double *r,*v2,*rv,**r3,rp,*rp3;
  double pom,pom1,pom2;
  double c2,delta;
  double gm;

  gm=ph[1];
  
  c2=DSQR(C1);
  r=dvector(0,ns);
  v2=dvector(0,ns);
  rv=dvector(0,ns);
  r3=dmatrix(0,ns,0,n);
  rp3=dvector(0,n);
  for(i=1;i<=ns;i++){
    r[i]=sqrt(DSQR(xs[i])+DSQR(ys[i])+DSQR(zs[i]));
    r3[i][0]=r[i]*r[i]*r[i];
    v2[i]=sqrt(DSQR(vxs[i])+DSQR(vys[i])+DSQR(vzs[i]));
    rv[i]=xs[i]*vxs[i]+ys[i]*vys[i]+zs[i]*vzs[i];
  }
  for(i=1;i<=ns;i++)
    for(l=1;l<=n;l++){
      delta=sqrt(DSQR(xs[i]-x[l])+DSQR(ys[i]-y[l])+DSQR(zs[i]-z[l]));
      r3[i][l]=delta*delta*delta;
    }
  for(l=1;l<=n;l++){
    rp=sqrt(DSQR(x[l])+DSQR(y[l])+DSQR(z[l]));
    rp3[l]=rp*rp*rp;
  }
  for(i=1;i<=ns;i++){
    pom=-gm/r3[i][0];
    axs[i]=pom*xs[i];
    ays[i]=pom*ys[i];
    azs[i]=pom*zs[i];
  }
  /* relativistic correction */
  pom=4*gm/c2;
  for(i=1;i<=ns;i++){
    pom1=gm/r3[i][0]*(pom/r[i]-v2[i]/c2);
    pom2=pom*rv[i]/r3[i][0];
    axs[i]+=pom1*xs[i]+pom2*vxs[i];
    ays[i]+=pom1*ys[i]+pom2*vys[i];
    azs[i]+=pom1*zs[i]+pom2*vzs[i];
  }
  for(i=1;i<=ns;i++)
    for(l=1;l<=n;l++){
      pom1=-gm*m[l]/r3[i][l];
      pom2=-gm*m[l]/rp3[l];
      axs[i]+=(pom1*(xs[i]-x[l])+pom2*x[l]);
      ays[i]+=(pom1*(ys[i]-y[l])+pom2*y[l]);
      azs[i]+=(pom1*(zs[i]-z[l])+pom2*z[l]);
    }
  free_dvector(r,0,ns);
  free_dvector(v2,0,ns);
  free_dvector(rv,0,ns);
  free_dmatrix(r3,0,ns,0,n);
  free_dvector(rp3,0,n);
}

void accsecvar(int n,double *m,double *x,double *y,double *z,int ns,
	       double *xs,double *ys,double *zs,
	       double *vxs,double *vys,double *vzs,
	       double *Xs,double *Ys,double *Zs,
	       double *Vxs,double *Vys,double *Vzs,
	       double *Axs,double *Ays,double *Azs,double *ph,int *flag)
{
  int i,l;
  double r,**r3,**r5;
  double pom,pom1,pom2;
  double gm;

  gm=ph[1];
  
  r3=dmatrix(0,ns,0,n);
  r5=dmatrix(0,ns,0,n);
  for(i=1;i<=ns;i++){
    if(flag[i]==1){
      r=sqrt(DSQR(xs[i])+DSQR(ys[i])+DSQR(zs[i]));
      r3[i][0]=r*r*r;
      r5[i][0]=r*r*r*r*r;
      for(l=1;l<=n;l++){
	r=sqrt(DSQR(xs[i]-x[l])+DSQR(ys[i]-y[l])+DSQR(zs[i]-z[l]));
	r3[i][l]=r*r*r;
	r5[i][l]=r*r*r*r*r;
      }
    }
  }
  for(i=1;i<=ns;i++){
    if(flag[i]==1){
      pom1=-gm/r3[i][0];
      pom2=3*gm/r5[i][0];
      Axs[i]=pom1*Xs[i]+pom2*(
			      Xs[i]*xs[i]*xs[i]+
			      Ys[i]*xs[i]*ys[i]+
			      Zs[i]*xs[i]*zs[i]);
      Ays[i]=pom1*Ys[i]+pom2*(
			      Xs[i]*ys[i]*xs[i]+
			      Ys[i]*ys[i]*ys[i]+
			      Zs[i]*ys[i]*zs[i]);
      Azs[i]=pom1*Zs[i]+pom2*(
			      Xs[i]*zs[i]*xs[i]+
			      Ys[i]*zs[i]*ys[i]+
			      Zs[i]*zs[i]*zs[i]);
      for(l=1;l<=n;l++){
	pom1=-gm*m[l]/r3[i][l];
	pom2=3.*gm*m[l]/r5[i][l];
	Axs[i]+=(pom1*Xs[i]+pom2*(
				  Xs[i]*(xs[i]-x[l])*(xs[i]-x[l])+
				  Ys[i]*(xs[i]-x[l])*(ys[i]-y[l])+
				  Zs[i]*(xs[i]-x[l])*(zs[i]-z[l])));
	Ays[i]+=(pom1*Ys[i]+pom2*(
				  Xs[i]*(ys[i]-y[l])*(xs[i]-x[l])+
				  Ys[i]*(ys[i]-y[l])*(ys[i]-y[l])+
				  Zs[i]*(ys[i]-y[l])*(zs[i]-z[l])));
	Azs[i]+=(pom1*Zs[i]+pom2*(
				  Xs[i]*(zs[i]-z[l])*(xs[i]-x[l])+
				  Ys[i]*(zs[i]-z[l])*(ys[i]-y[l])+
				  Zs[i]*(zs[i]-z[l])*(zs[i]-z[l])));
      }
    } else {
      Axs[i]=Ays[i]=Azs[i]=0;
    }
  }
  free_dmatrix(r3,0,ns,0,n);
  free_dmatrix(r5,0,ns,0,n);
}

void asvrel(int n,double *m,double *x,double *y,double *z,int ns,
	    double *xs,double *ys,double *zs,
	    double *vxs,double *vys,double *vzs,
	    double *Xs,double *Ys,double *Zs,
	    double *Vxs,double *Vys,double *Vzs,
	    double *Axs,double *Ays,double *Azs,double *ph,int *flag)
{
  int i,l;
  double *r,**r3,**r5;
  double pom,pom1,pom2;
  double c2;
  double v2,rR,vV,vR,rv,rV;
  double gm;

  gm=ph[1];
  
  r=dvector(0,ns);
  r3=dmatrix(0,ns,0,n);
  r5=dmatrix(0,ns,0,n);
  c2=DSQR(C1);
  for(i=1;i<=ns;i++){
    pom=sqrt(DSQR(xs[i])+DSQR(ys[i])+DSQR(zs[i]));
    r[i]=pom;
    r3[i][0]=pom*pom*pom;
    r5[i][0]=pom*pom*pom*pom*pom;
  }
  for(i=1;i<=ns;i++)
    for(l=1;l<=n;l++){
      pom=sqrt(DSQR(xs[i]-x[l])+DSQR(ys[i]-y[l])+DSQR(zs[i]-z[l]));
      r3[i][l]=pom*pom*pom;
      r5[i][l]=pom*pom*pom*pom*pom;
    }
  for(i=1;i<=ns;i++){
    pom1=-gm/r3[i][0];
    pom2=3*gm/r5[i][0];
    rR=xs[i]*Xs[i]+ys[i]*Ys[i]+zs[i]*Zs[i];
    Axs[i]=pom1*Xs[i]+pom2*xs[i]*rR;
    Ays[i]=pom1*Ys[i]+pom2*ys[i]*rR;
    Azs[i]=pom1*Zs[i]+pom2*zs[i]*rR;
  }
  /* relativistic correction */
  pom=gm/c2;
  for(i=1;i<=ns;i++){
    pom1=4*gm*pom/r[i];
    v2=vxs[i]*vxs[i]+vys[i]*vys[i]+vzs[i]*vzs[i];
    rv=xs[i]*vxs[i]+ys[i]*vys[i]+zs[i]*vzs[i];
    rR=xs[i]*Xs[i]+ys[i]*Ys[i]+zs[i]*Zs[i];
    rV=xs[i]*Vxs[i]+ys[i]*Vys[i]+zs[i]*Vzs[i];
    vR=vxs[i]*Xs[i]+vys[i]*Ys[i]+vzs[i]*Zs[i];
    vV=vxs[i]*Vxs[i]+vys[i]*Vys[i]+vzs[i]*Vzs[i];
    Axs[i]+=pom1*(Xs[i]/r3[i][0]-4*rR*xs[i]/r5[i][0])+
      pom*(-v2*Xs[i]/r3[i][0]+3*v2*rR*xs[i]/r5[i][0]-2*vV*xs[i]/r3[i][0])+
	4*pom*(vR*vxs[i]/r3[i][0]-3*rv*rR*vxs[i]/r5[i][0]+rv*Vxs[i]/r3[i][0]+
	       rV*vxs[i]/r3[i][0]);
    Ays[i]+=pom1*(Ys[i]/r3[i][0]-4*rR*ys[i]/r5[i][0])+
      pom*(-v2*Ys[i]/r3[i][0]+3*v2*rR*ys[i]/r5[i][0]-2*vV*ys[i]/r3[i][0])+
	4*pom*(vR*vys[i]/r3[i][0]-3*rv*rR*vys[i]/r5[i][0]+rv*Vys[i]/r3[i][0]+
	       rV*vys[i]/r3[i][0]);
    Azs[i]+=pom1*(Zs[i]/r3[i][0]-4*rR*zs[i]/r5[i][0])+
      pom*(-v2*Zs[i]/r3[i][0]+3*v2*rR*zs[i]/r5[i][0]-2*vV*zs[i]/r3[i][0])+
	4*pom*(vR*vzs[i]/r3[i][0]-3*rv*rR*vzs[i]/r5[i][0]+rv*Vzs[i]/r3[i][0]+
	       rV*vzs[i]/r3[i][0]);
  }
/* ================================================================ */
  for(i=1;i<=ns;i++)
    for(l=1;l<=n;l++){
      pom=Xs[i]*(xs[i]-x[l])+Ys[i]*(ys[i]-y[l])+Zs[i]*(zs[i]-z[l]);
      pom1=-gm*m[l]/r3[i][l];
      pom2=3.*gm*m[l]/r5[i][l];
      Axs[i]+=pom1*Xs[i]+pom2*pom*(xs[i]-x[l]);
      Ays[i]+=pom1*Ys[i]+pom2*pom*(ys[i]-y[l]);
      Azs[i]+=pom1*Zs[i]+pom2*pom*(zs[i]-z[l]);
    }
  free_dvector(r,0,ns);
  free_dmatrix(r3,0,ns,0,n);
  free_dmatrix(r5,0,ns,0,n);
}

void accsec2(int n,double *m,double *x,double *y,double *z,
	     int ns,double *xs,double *ys,double *zs,
	     double *vxs,double *vys,double *vzs,
	     double *axs,double *ays,double *azs,double *ph)
     /* Stokes drag */
{
  int i,l;
  double r,**r3,rp,*rp3,*r32,*vr;
  double pom,pom1,pom2,om;
  double gm,draga,dragc;

  gm=ph[1];
  draga=ph[2];
  dragc=ph[3];
  
  r3=dmatrix(0,ns,0,n);
  rp3=dvector(0,n);
  r32=dvector(0,ns);
  for(i=1;i<=ns;i++){
    r=sqrt(DSQR(xs[i])+DSQR(ys[i])+DSQR(zs[i]));
    r3[i][0]=r*r*r;
    r32[i]=sqrt(r3[i][0]);
  }
  for(i=1;i<=ns;i++)
    for(l=1;l<=n;l++){
      r=sqrt(DSQR(xs[i]-x[l])+DSQR(ys[i]-y[l])+DSQR(zs[i]-z[l]));
      r3[i][l]=r*r*r;
    }
  for(l=1;l<=n;l++){
    rp=sqrt(DSQR(x[l])+DSQR(y[l])+DSQR(z[l]));
    rp3[l]=rp*rp*rp;
  }
  for(i=1;i<=ns;i++){
    pom=-gm/r3[i][0];
    axs[i]=pom*xs[i];
    ays[i]=pom*ys[i];
    azs[i]=pom*zs[i];
  }
  for(i=1;i<=ns;i++)
    for(l=1;l<=n;l++){
      pom1=-gm*m[l]/r3[i][l];
      pom2=-gm*m[l]/rp3[l];
      axs[i]+=(pom1*(xs[i]-x[l])+pom2*x[l]);
      ays[i]+=(pom1*(ys[i]-y[l])+pom2*y[l]);
      azs[i]+=(pom1*(zs[i]-z[l])+pom2*z[l]);
    }
  for(i=1;i<=ns;i++){
    om=sqrt(gm)/r32[i];
    axs[i]+=-dragc*(vxs[i]-draga*(-om*ys[i]));
    ays[i]+=-dragc*(vys[i]-draga*(om*xs[i]));
    azs[i]+=-dragc*vzs[i];
  }
  free_dmatrix(r3,0,ns,0,n);
  free_dvector(rp3,0,n);
  free_dvector(r32,0,ns);
}

void gomsi(int p, double **x, double **y, double **z,
	   double **ax, double **ay, double **az,
	   int n, int k, double *alfa, double *beta, double h)
{
  int i,j,l,m;
  double h2;
  
  h2=h*h;
  l=n+k;
  for(i=1;i<=p;i++) {                 /* oldest point */
    x[l][i]=h2*beta[0]*ax[n][i]-alfa[0]*x[n][i];
    y[l][i]=h2*beta[0]*ay[n][i]-alfa[0]*y[n][i];
    z[l][i]=h2*beta[0]*az[n][i]-alfa[0]*z[n][i];
  }
  for(j=1;j<=k-1;j++){      /* integrator is explicit */
    m=n+j;
    for(i=1;i<=p;i++){
      x[l][i]+=(h2*beta[j]*ax[m][i]-alfa[j]*x[m][i]);
      y[l][i]+=(h2*beta[j]*ay[m][i]-alfa[j]*y[m][i]);
      z[l][i]+=(h2*beta[j]*az[m][i]-alfa[j]*z[m][i]);
    }
  }
}

void gork(int p,double *m,double **x,double **y,double **z,
	  double **vx,double **vy,double **vz,
	  int ps,double **xs,double **ys,double **zs,
	  double **vxs,double **vys,double **vzs,
	  int n,int k,double *a,double *q,double h,
	  double *ph,int *flag,
	  void (*f1)(int,double *,double *,double *,double *,
		     double *,double *,double *,double *,double *,
		     double *,double *),
	  void (*f2)(int, double *,double *,double *,double *,
		     int, double *,double *, double *,
		     double *,double *,double *,double *,double *,
		     double *,double *,int *))
{
  double *x2,*y2,*z2,*vx2,*vy2,*vz2;
  double *kx,*ky,*kz,*kvx,*kvy,*kvz;
  double *ax,*ay,*az;
  double *xs2,*ys2,*zs2,*vxs2,*vys2,*vzs2;
  double *kxs,*kys,*kzs,*kvxs,*kvys,*kvzs;
  double *axs,*ays,*azs;
  double pom;
  int i,j,l;
/* ======================================================== */
  x2=dvector(0,p); y2=dvector(0,p); z2=dvector(0,p);
  vx2=dvector(0,p); vy2=dvector(0,p); vz2=dvector(0,p);
  kx=dvector(0,p); ky=dvector(0,p); kz=dvector(0,p);
  kvx=dvector(0,p); kvy=dvector(0,p); kvz=dvector(0,p);
  ax=dvector(0,p); ay=dvector(0,p); az=dvector(0,p);

  xs2=dvector(0,ps); ys2=dvector(0,ps); zs2=dvector(0,ps);
  vxs2=dvector(0,ps);	vys2=dvector(0,ps); vzs2=dvector(0,ps);
  kxs=dvector(0,ps); kys=dvector(0,ps); kzs=dvector(0,ps);
  kvxs=dvector(0,ps); kvys=dvector(0,ps); kvzs=dvector(0,ps);
  axs=dvector(0,ps); ays=dvector(0,ps); azs=dvector(0,ps);
/* ========================================================= */
  for(i=0;i<=p;i++){
    x2[i]=0;y2[i]=0;z2[i]=0;vx2[i]=0;vy2[i]=0;vz2[i]=0;
    kx[i]=0;ky[i]=0;kz[i]=0;kvx[i]=0;kvy[i]=0;kvz[i]=0;
  }
  for(i=0;i<=ps;i++){
    xs2[i]=0;ys2[i]=0;zs2[i]=0;vxs2[i]=0;vys2[i]=0;vzs2[i]=0;
    kxs[i]=0;kys[i]=0;kzs[i]=0;kvxs[i]=0;kvys[i]=0;kvzs[i]=0;
  }
  for(j=1;j<5;j++){
    pom=q[j]*h;
    for(i=1;i<=p;i++){
      x2[i]=x[k][i]+pom*kx[i];
      y2[i]=y[k][i]+pom*ky[i];
      z2[i]=z[k][i]+pom*kz[i];
      vx2[i]=vx[k][i]+pom*kvx[i];
      vy2[i]=vy[k][i]+pom*kvy[i];
      vz2[i]=vz[k][i]+pom*kvz[i];
    }
    for(i=1;i<=ps;i++){
      xs2[i]=xs[k][i]+pom*kxs[i];
      ys2[i]=ys[k][i]+pom*kys[i];
      zs2[i]=zs[k][i]+pom*kzs[i];
      vxs2[i]=vxs[k][i]+pom*kvxs[i];
      vys2[i]=vys[k][i]+pom*kvys[i];
      vzs2[i]=vzs[k][i]+pom*kvzs[i];
    }
    (*f1)(p,m,x2,y2,z2,vx2,vy2,vz2,ax,ay,az,ph);
    (*f2)(p,m,x2,y2,z2,ps,xs2,ys2,zs2,vxs2,vys2,vzs2,
	  axs,ays,azs,ph,flag);
    for(i=1;i<=p;i++){
      kx[i]=vx2[i];
      ky[i]=vy2[i];
      kz[i]=vz2[i];
      kvx[i]=ax[i];
      kvy[i]=ay[i];
      kvz[i]=az[i];
    }
    for(i=1;i<=ps;i++){
      kxs[i]=vxs2[i];
      kys[i]=vys2[i];
      kzs[i]=vzs2[i];
      kvxs[i]=axs[i];
      kvys[i]=ays[i];
      kvzs[i]=azs[i];
    }
    for(i=1;i<=p;i++){
      x[n][i]+=h*a[j]*kx[i];
      y[n][i]+=h*a[j]*ky[i];
      z[n][i]+=h*a[j]*kz[i];
      vx[n][i]+=h*a[j]*kvx[i];
      vy[n][i]+=h*a[j]*kvy[i];
      vz[n][i]+=h*a[j]*kvz[i];
    }
    for(i=1;i<=ps;i++){
      xs[n][i]+=h*a[j]*kxs[i];
      ys[n][i]+=h*a[j]*kys[i];
      zs[n][i]+=h*a[j]*kzs[i];
      vxs[n][i]+=h*a[j]*kvxs[i];
      vys[n][i]+=h*a[j]*kvys[i];
      vzs[n][i]+=h*a[j]*kvzs[i];
    }
  }
/* ===================================================== */
  free_dvector(x2,0,p); free_dvector(y2,0,p); free_dvector(z2,0,p);
  free_dvector(vx2,0,p); free_dvector(vy2,0,p); free_dvector(vz2,0,p);
  free_dvector(kx,0,p); free_dvector(ky,0,p); free_dvector(kz,0,p);
  free_dvector(kvx,0,p); free_dvector(kvy,0,p); free_dvector(kvz,0,p);
  free_dvector(ax,0,p); free_dvector(ay,0,p); free_dvector(az,0,p);

  free_dvector(xs2,0,ps); free_dvector(ys2,0,ps); free_dvector(zs2,0,ps);
  free_dvector(vxs2,0,ps); free_dvector(vys2,0,ps); free_dvector(vzs2,0,ps);
  free_dvector(kxs,0,ps); free_dvector(kys,0,ps); free_dvector(kzs,0,ps);
  free_dvector(kvxs,0,ps); free_dvector(kvys,0,ps); free_dvector(kvzs,0,ps);
  free_dvector(axs,0,ps); free_dvector(ays,0,ps); free_dvector(azs,0,ps);
}

void gorkvar(int p,double *m,double **x,double **y,double **z,
	     double **vx,double **vy,double **vz,int ps,
	     double **xs,double **ys,double **zs,
	     double **vxs,double **vys,double **vzs,
	     double **Xs,double **Ys,double **Zs,
	     double **Vxs,double **Vys,double **Vzs,
	     int n,int k,double *a,double *q,double h,
	     double *ph,int *flag,
	     void (*f1)(int,double *,double *,double *,double *,
			double *,double *,double *,double *,double *,double *,
			double *),
	     void (*f2)(int, double *,double *,double *,double *,
			int, double *,double *, double *,
			double *,double *,double *,double *,double *,double *,
			double *,int *),
	     void (*f3)(int, double *,double *,double *,double *,int,
			double *,double *,double *,double *,double *,double *,
			double *,double *,double *,double *,double *,double *,
			double *,double *,double *,
			double *,int *))
{
  double *x2,*y2,*z2,*vx2,*vy2,*vz2;
  double *kx,*ky,*kz,*kvx,*kvy,*kvz;
  double *ax,*ay,*az;
  double *xs2,*ys2,*zs2,*vxs2,*vys2,*vzs2;
  double *kxs,*kys,*kzs,*kvxs,*kvys,*kvzs;
  double *axs,*ays,*azs;
  double *Xs2,*Ys2,*Zs2,*Vxs2,*Vys2,*Vzs2;
  double *Kxs,*Kys,*Kzs,*Kvxs,*Kvys,*Kvzs;
  double *Axs,*Ays,*Azs;
  double pom;
  int i,j,l;
/* ======================================================== */
  x2=dvector(0,p); y2=dvector(0,p); z2=dvector(0,p);
  vx2=dvector(0,p); vy2=dvector(0,p); vz2=dvector(0,p);
  kx=dvector(0,p); ky=dvector(0,p); kz=dvector(0,p);
  kvx=dvector(0,p); kvy=dvector(0,p); kvz=dvector(0,p);
  ax=dvector(0,p); ay=dvector(0,p); az=dvector(0,p);

  xs2=dvector(0,ps); ys2=dvector(0,ps); zs2=dvector(0,ps);
  vxs2=dvector(0,ps); vys2=dvector(0,ps); vzs2=dvector(0,ps);
  kxs=dvector(0,ps); kys=dvector(0,ps); kzs=dvector(0,ps);
  kvxs=dvector(0,ps); kvys=dvector(0,ps); kvzs=dvector(0,ps);
  axs=dvector(0,ps); ays=dvector(0,ps); azs=dvector(0,ps);

  Xs2=dvector(0,ps); Ys2=dvector(0,ps); Zs2=dvector(0,ps);
  Vxs2=dvector(0,ps); Vys2=dvector(0,ps); Vzs2=dvector(0,ps);
  Kxs=dvector(0,ps); Kys=dvector(0,ps); Kzs=dvector(0,ps);
  Kvxs=dvector(0,ps); Kvys=dvector(0,ps); Kvzs=dvector(0,ps);
  Axs=dvector(0,ps); Ays=dvector(0,ps); Azs=dvector(0,ps);
/* ========================================================= */
  for(i=0;i<=p;i++){
    x2[i]=0;y2[i]=0;z2[i]=0;vx2[i]=0;vy2[i]=0;vz2[i]=0;
    kx[i]=0;ky[i]=0;kz[i]=0;kvx[i]=0;kvy[i]=0;kvz[i]=0;
  }
  for(i=0;i<=ps;i++){
    xs2[i]=0;ys2[i]=0;zs2[i]=0;vxs2[i]=0;vys2[i]=0;vzs2[i]=0;
    kxs[i]=0;kys[i]=0;kzs[i]=0;kvxs[i]=0;kvys[i]=0;kvzs[i]=0;
  }
  for(i=0;i<=ps;i++){
    Xs2[i]=0;Ys2[i]=0;Zs2[i]=0;Vxs2[i]=0;Vys2[i]=0;Vzs2[i]=0;
    Kxs[i]=0;Kys[i]=0;Kzs[i]=0;Kvxs[i]=0;Kvys[i]=0;Kvzs[i]=0;
  }
  for(j=1;j<5;j++){
    pom=q[j]*h;
    for(i=1;i<=p;i++){
      x2[i]=x[k][i]+pom*kx[i];
      y2[i]=y[k][i]+pom*ky[i];
      z2[i]=z[k][i]+pom*kz[i];
      vx2[i]=vx[k][i]+pom*kvx[i];
      vy2[i]=vy[k][i]+pom*kvy[i];
      vz2[i]=vz[k][i]+pom*kvz[i];
    }
    for(i=1;i<=ps;i++){
      xs2[i]=xs[k][i]+pom*kxs[i];
      ys2[i]=ys[k][i]+pom*kys[i];
      zs2[i]=zs[k][i]+pom*kzs[i];
      vxs2[i]=vxs[k][i]+pom*kvxs[i];
      vys2[i]=vys[k][i]+pom*kvys[i];
      vzs2[i]=vzs[k][i]+pom*kvzs[i];
    }
    for(i=1;i<=ps;i++){
      Xs2[i]=Xs[k][i]+pom*Kxs[i];
      Ys2[i]=Ys[k][i]+pom*Kys[i];
      Zs2[i]=Zs[k][i]+pom*Kzs[i];
      Vxs2[i]=Vxs[k][i]+pom*Kvxs[i];
      Vys2[i]=Vys[k][i]+pom*Kvys[i];
      Vzs2[i]=Vzs[k][i]+pom*Kvzs[i];
    }
    (*f1)(p,m,x2,y2,z2,vx2,vy2,vz2,ax,ay,az,ph);
    (*f2)(p,m,x2,y2,z2,ps,xs2,ys2,zs2,vxs2,vys2,vzs2,axs,ays,azs,ph,flag);
    (*f3)(p,m,x2,y2,z2,ps,xs2,ys2,zs2,vxs2,vys2,vzs2,Xs2,Ys2,Zs2,
	  Vxs2,Vys2,Vzs2,Axs,Ays,Azs,ph,flag);
    for(i=1;i<=p;i++){
      kx[i]=vx2[i];
      ky[i]=vy2[i];
      kz[i]=vz2[i];
      kvx[i]=ax[i];
      kvy[i]=ay[i];
      kvz[i]=az[i];
    }
    for(i=1;i<=ps;i++){
      kxs[i]=vxs2[i];
      kys[i]=vys2[i];
      kzs[i]=vzs2[i];
      kvxs[i]=axs[i];
      kvys[i]=ays[i];
      kvzs[i]=azs[i];
    }
    for(i=1;i<=ps;i++){
      Kxs[i]=Vxs2[i];
      Kys[i]=Vys2[i];
      Kzs[i]=Vzs2[i];
      Kvxs[i]=Axs[i];
      Kvys[i]=Ays[i];
      Kvzs[i]=Azs[i];
    }
    for(i=1;i<=p;i++){
      x[n][i]+=h*a[j]*kx[i];
      y[n][i]+=h*a[j]*ky[i];
      z[n][i]+=h*a[j]*kz[i];
      vx[n][i]+=h*a[j]*kvx[i];
      vy[n][i]+=h*a[j]*kvy[i];
      vz[n][i]+=h*a[j]*kvz[i];
    }
    for(i=1;i<=ps;i++){
      xs[n][i]+=h*a[j]*kxs[i];
      ys[n][i]+=h*a[j]*kys[i];
      zs[n][i]+=h*a[j]*kzs[i];
      vxs[n][i]+=h*a[j]*kvxs[i];
      vys[n][i]+=h*a[j]*kvys[i];
      vzs[n][i]+=h*a[j]*kvzs[i];
    }
    for(i=1;i<=ps;i++){
      Xs[n][i]+=h*a[j]*Kxs[i];
      Ys[n][i]+=h*a[j]*Kys[i];
      Zs[n][i]+=h*a[j]*Kzs[i];
      Vxs[n][i]+=h*a[j]*Kvxs[i];
      Vys[n][i]+=h*a[j]*Kvys[i];
      Vzs[n][i]+=h*a[j]*Kvzs[i];
    }
  }
/* ===================================================== */
  free_dvector(x2,0,p); free_dvector(y2,0,p); free_dvector(z2,0,p);
  free_dvector(vx2,0,p); free_dvector(vy2,0,p); free_dvector(vz2,0,p);
  free_dvector(kx,0,p); free_dvector(ky,0,p); free_dvector(kz,0,p);
  free_dvector(kvx,0,p); free_dvector(kvy,0,p); free_dvector(kvz,0,p);
  free_dvector(ax,0,p); free_dvector(ay,0,p); free_dvector(az,0,p);

  free_dvector(xs2,0,ps); free_dvector(ys2,0,ps); free_dvector(zs2,0,ps);
  free_dvector(vxs2,0,ps); free_dvector(vys2,0,ps); free_dvector(vzs2,0,ps);
  free_dvector(kxs,0,ps); free_dvector(kys,0,ps); free_dvector(kzs,0,ps);
  free_dvector(kvxs,0,ps); free_dvector(kvys,0,ps); free_dvector(kvzs,0,ps);
  free_dvector(axs,0,ps); free_dvector(ays,0,ps); free_dvector(azs,0,ps);

  free_dvector(Xs2,0,ps); free_dvector(Ys2,0,ps); free_dvector(Zs2,0,ps);
  free_dvector(Vxs2,0,ps); free_dvector(Vys2,0,ps); free_dvector(Vzs2,0,ps);
  free_dvector(Kxs,0,ps); free_dvector(Kys,0,ps); free_dvector(Kzs,0,ps);
  free_dvector(Kvxs,0,ps); free_dvector(Kvys,0,ps); free_dvector(Kvzs,0,ps);
  free_dvector(Axs,0,ps); free_dvector(Ays,0,ps); free_dvector(Azs,0,ps);
}

void katos(int p,double *m,double t,double *a,double *e,double *ii,
	   double *o,double *oo,double *nt,double *x,double *y,double *z,
	   double *vx,double *vy,double *vz,double *ph)
{
  int i;
  double k,r,v,ee,mx,my,mz,mom,u,xu,yu,zu,xp,yp,zp;
  double pom,pom1,pom2,pom3,pom4;
  double gm;

  gm=ph[1];
  
  for(i=1;i<=p;i++){
    k=gm*(1+m[i]);
    r=sqrt(x[i]*x[i]+y[i]*y[i]+z[i]*z[i]);
    v=sqrt(vx[i]*vx[i]+vy[i]*vy[i]+vz[i]*vz[i]);
    ee=v*v/2.0-k/r;
    if(ee > 0.0) nrerror("energy is positive\n");
    mx=(y[i]*vz[i]-z[i]*vy[i]);
    my=(z[i]*vx[i]-x[i]*vz[i]);
    mz=(x[i]*vy[i]-y[i]*vx[i]);
    mom=sqrt(mx*mx+my*my+mz*mz);
    a[i]=-k/(2.0*ee);
    pom1=mz/mom;
    
    if(fabs(pom1) < 1.0) {      
/* for ii < 1e-6 inaccurate */
      ii[i]=acos(pom1);
      if(my != 0.0){
	if((oo[i]=atan2(mx,-my)) < 0.0) oo[i]+=PI2;
      } else {
	if(mx > 0.0) oo[i]=PI2/4.0;
	else oo[i]=3.0*PI2/4.0;
      }
    } else {
      ii[i]=0.0;
      oo[i]=0.0;
    }
    
    if((pom1=1.0-mom*mom/(k*a[i])) > 1e-15) {  
/* for e < 1e-6 inaccurate */
      e[i]=sqrt(pom1);
      pom=k*e[i];
      xp=(vy[i]*mz-vz[i]*my)/pom-x[i]/(r*e[i]);
      yp=(vz[i]*mx-vx[i]*mz)/pom-y[i]/(r*e[i]);
      zp=(vx[i]*my-vy[i]*mx)/pom-z[i]/(r*e[i]);
      xu=cos(oo[i]);yu=sin(oo[i]);zu=0.0;
      pom2=xp*xu+yp*yu+zp*zu;
      if(fabs(pom2) <= 1.0)  o[i]=acos(pom2); 
/* very near to perihelium an error 1e-5 deg, possible to correct */
      else  o[i]=0.0;
      if((zp==0.0 && yp < 0.0) || zp < 0.0) o[i]+=PI2-2*o[i];
      pom1=1-r/a[i];
      pom2=(x[i]*vx[i]+y[i]*vy[i]+z[i]*vz[i])/sqrt(k*a[i]);
      if(pom1 != 0.0) {
	if((u=atan2(pom2,pom1)) < 0.0) u+=PI2;
      } else {
	if(pom2 > 0.0) u=PI2/4.0;
	else u=3.0*PI2/4.0;
      }
      if((nt[i]=u-e[i]*sin(u)) < 0.0) nt[i]+=PI2;
    } else {
      e[i]=0.0;
      o[i]=0.0;
      pom1=(x[i]*cos(oo[i])+y[i]*sin(oo[i]))/r;
      
      if(fabs(pom1) <= 1.0) nt[i]=acos(pom1);
      else nt[i]=0.0;
      
      if((z[i]==0.0 && y[i]<0.0) || z[i]<0.0) nt[i]+=PI2-2*nt[i];
    }
  }
}

void katos2(int p,double *m,double t,double *a,double *e,double *ii,
	    double *o,double *oo,double *nt,double *x,double *y,double *z,
	    double *vx,double *vy,double *vz,int *flag,double *ph)
{
  int i;
  double k,r,v,ee,mx,my,mz,mom,u,xu,yu,zu,xp,yp,zp;
  double pom,pom1,pom2,pom3,pom4;
  double gm;

  gm=ph[1];
  
  t=t;
  for(i=1;i<=p;i++){
    if(flag[i]==0){
       a[i]=e[i]=ii[i]=o[i]=oo[i]=nt[i]=0;
    } else {  
      k=gm*(1+m[i]);
      r=sqrt(x[i]*x[i]+y[i]*y[i]+z[i]*z[i]);
      v=sqrt(vx[i]*vx[i]+vy[i]*vy[i]+vz[i]*vz[i]);
      ee=v*v/2.0-k/r;
      mx=(y[i]*vz[i]-z[i]*vy[i]);
      my=(z[i]*vx[i]-x[i]*vz[i]);
      mz=(x[i]*vy[i]-y[i]*vx[i]);
      mom=sqrt(mx*mx+my*my+mz*mz);
      if(ee >= 0.0 || mom==0.0) {
	flag[i]=0;
	a[i]=0;e[i]=0;ii[i]=0;o[i]=0;oo[i]=0;nt[i]=0;
      } else {
	a[i]=-k/(2.0*ee);
	pom1=mz/mom;
	
	if(fabs(pom1) < 1.0) {
	  /* for ii < 1e-6 inaccurate */
	  ii[i]=acos(pom1);
	  if(my != 0.0){
	    if((oo[i]=atan2(mx,-my)) < 0.0) oo[i]+=PI2;
	  } else {
	    if(mx > 0.0) oo[i]=PI2/4.0;
	    else oo[i]=3.0*PI2/4.0;
	  }
	} else {
	  ii[i]=0.0;
	  oo[i]=0.0;
	}
	
	if((pom1=1.0-mom*mom/(k*a[i])) > 1e-15) {
	  /* for e < 1e-6 inaccurate */
	  e[i]=sqrt(pom1);
	  pom=k*e[i];
	  xp=(vy[i]*mz-vz[i]*my)/pom-x[i]/(r*e[i]);
	  yp=(vz[i]*mx-vx[i]*mz)/pom-y[i]/(r*e[i]);
	  zp=(vx[i]*my-vy[i]*mx)/pom-z[i]/(r*e[i]);
	  xu=cos(oo[i]);yu=sin(oo[i]);zu=0.0;
	  pom2=xp*xu+yp*yu+zp*zu;
	  if(fabs(pom2) <= 1.0)  o[i]=acos(pom2);
	  else  o[i]=0.0;
	  if((zp==0.0 && yp < 0.0) || zp < 0.0) o[i]+=PI2-2*o[i];
	  pom1=1-r/a[i];
	  pom2=(x[i]*vx[i]+y[i]*vy[i]+z[i]*vz[i])/sqrt(k*a[i]);
	  if(pom1 != 0.0) {
	    if((u=atan2(pom2,pom1)) < 0.0) u+=PI2;
	  } else {
	    if(pom2 > 0.0) u=PI2/4.0;
	    else u=3.0*PI2/4.0;
	  }
	  if((nt[i]=u-e[i]*sin(u)) < 0.0) nt[i]+=PI2;
	} else {
	  e[i]=0.0;
	  o[i]=0.0;
	  pom1=(x[i]*cos(oo[i])+y[i]*sin(oo[i]))/r;
	  
	  if(fabs(pom1) <= 1.0) nt[i]=acos(pom1);
	  else nt[i]=0.0;
	  
	  if((z[i]==0.0 && y[i]<0.0) || z[i]<0.0) nt[i]+=PI2-2*nt[i];
	}
      }
    }
  }
}  

void ostoka(int p,double *m,double t,double *a,double *e,double *ii,
	    double *o,double *oo,double *nt,double *x,double *y,double *z,
	    double *vx,double *vy,double *vz,double *ph)
{
  int i,l;
  double k,x2,y2,z2,vx2,vy2,vz2,sdp,delta,u,ud;
  double co,coo,cii,so,soo,sii,rot[4][4];
  double gm;

  gm=ph[1];
  
  for(i=1;i<=p;i++) {
    /* k=gm*(1+m[i]); */
    k = gm; 
    sdp=sqrt(k/a[i])/a[i];
    u=nt[i];
    l=1;
    while(fabs(delta=(u-e[i]*sin(u)-nt[i])/(1-e[i]*cos(u))) > 1e-13 && l<100){
      u=u-delta;
      l++;
    }
    if(l==100){}
    x2=a[i]*(cos(u)-e[i]);
    y2=a[i]*sqrt(1-e[i]*e[i])*sin(u);
    z2=0;
    ud=sdp/(1-e[i]*cos(u));
    vx2=-a[i]*sin(u)*ud;
    vy2=a[i]*sqrt(1-e[i]*e[i])*cos(u)*ud;
    vz2=0;
    co=cos(o[i]);so=sin(o[i]);
    coo=cos(oo[i]);soo=sin(oo[i]);
    cii=cos(ii[i]);sii=sin(ii[i]);
    rot[1][1]=coo*co-soo*cii*so;
    rot[1][2]=-(coo*so+soo*cii*co);
    rot[1][3]=0;
    rot[2][1]=soo*co+coo*cii*so;
    rot[2][2]=-soo*so+coo*cii*co;
    rot[2][3]=0;
    rot[3][1]=sii*so;
    rot[3][2]=sii*co;
    rot[3][3]=0;
    multimat(rot[1],rot[2],rot[3],x2,y2,z2,&x[i],&y[i],&z[i]);
    multimat(rot[1],rot[2],rot[3],vx2,vy2,vz2,&vx[i],&vy[i],&vz[i]);
  }
}

void vel(int p, double **x, double **y, double **z,
	 double *vx, double *vy, double *vz,
	 double **ax, double **ay, double **az,
	 int n, int k, double *alfa, double *beta, double h)
{
  int i,j,l,m;
  double h2;
  
  h2=h*h;
  for(i=1;i<=p;i++) {                
    vx[i]=h2*beta[0]*ax[n][i]-alfa[0]*x[n][i];
    vy[i]=h2*beta[0]*ay[n][i]-alfa[0]*y[n][i];
    vz[i]=h2*beta[0]*az[n][i]-alfa[0]*z[n][i];
  }
  
  for(j=1;j<=k-1;j++){     
    m=n+j;
    for(i=1;i<=p;i++){
      vx[i]+=(h2*beta[j]*ax[m][i]-alfa[j]*x[m][i]);
      vy[i]+=(h2*beta[j]*ay[m][i]-alfa[j]*y[m][i]);
      vz[i]+=(h2*beta[j]*az[m][i]-alfa[j]*z[m][i]);
    }
  }
  for(i=1;i<=p;i++){
    vx[i]=vx[i]/h;
    vy[i]=vy[i]/h;
    vz[i]=vz[i]/h;
  }
}

